#!/usr/bin/perl -w -T
########################################################
#
# This script is intended to provide a means for
# detecting changes made to files, via a regular
# comparison of MD5 hashes to an established
# "baseline".
# In this respect, it is designed as a portable clone
# of aide software.
#
# This script requires perl ,and some others perl modules
# which come in standard installation
#
###############################################################################
#    Copyright (C) 2002 by Eric Gerbier
#    Bug reports to: gerbier@users.sourceforge.net
#    $Id: afick.pl,v 1.42 2004/01/08 09:17:33 gerbier Exp $
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
###############################################################################

use strict;

use Digest::MD5;    # for md5 checksum

# use Digest::SHA1;	# be careful : does not exist in standard on linux

use SDBM_File;       # database
use FileHandle;      # for database open mode
use Getopt::Long;    # arg analysis

#use Pod::Usage;
use File::Glob ':glob';    # for jokers
use File::Basename;        # for delete

################# global variables #############################

use vars qw ($Dbm);

$Dbm = 'SDBM_File';        # database type

use vars
  qw (%Finded %Hashfile %Newfiles %Dangling %Toscan %Onlydir %Onlythis %Oldval %Newval @ExSufx);
%Finded   = ();            # list of files scanned
%Hashfile = ();            # dbm hash-table
%Newfiles = ();            # list of new files
%Dangling = ();            # dangling links
%Toscan   = ();            # list to scan
%Onlydir  = ();            # only this and content, without sub-dir
%Onlythis = ();            # only this inode
%Oldval   = ();            # store old values
%Newval   = ();            # store new changes values
@ExSufx   = ();            # store excluded suffixes

use vars qw ( $Nbmod  $Update $File);
$Nbmod  = 0;               # number of modifications
$Update = 0;               # update/compare mode ?
$File   = 0;               # file mode ?

# directives status
# not initialized to be able to detect negative command lines
use vars
  qw ($Verbose $Warn_dead_symlinks $Report_full_newdel $Warn_missing_file $Report_url $Progress $Timing $Running);

#$Verbose            ;           # verbose mode ?
#$Warn_dead_symlinks = '';           # warn_dead_symlink
#$Report_full_newdel     = '';	   # report new directories contents
#$Warn_missing_file   = '';
$Report_url = "STDOUT";    # report_url

# alias table
# contains pre-defined values
my %alias = (
	a   => "a",              # atime
	b   => "b",              # blocks
	c   => "c",              # ctime
	d   => "d",              # device
	g   => "g",              # gid
	i   => "i",              # inode
	m   => "m",              # mtime
	md5 => "5",              # md5 hash
	h   => "5",              # md5 hash by default (compatibility)
	n   => "n",              # number of links
	p   => "p",              # permissions
	u   => "u",              # uid
	s   => "s",              # size
	all => "pdinugsbmc5",    # all except a
	R   => "pdinugsmc5",     # aide compatibility
	L   => "pdinug",         # aide compatibility
	E   => "",               # aide compatibility
);
################ end of global var #############################

my @listerep;                # list of rep to scan
my $database;                # database name

# command line arguments/flags
my $configfile;              # config file name
my $help;
my $init;
my $compare;
my $list;
my $print;
my $Version;
my $Sufx_list;
my $update;

my $database_ext = ".pag";    # database extension
my $control_ext  = ".ctr";    # database extension

my $version = "1.6-0";

my $default_config_file;
my $default_alias = "all";

#######################################################
# get checksum from database
sub get_md5 ($) {
	my $name = shift (@_);    # file name

	my $hash = $Hashfile{$name} || $Newval{$name} || "";
	my @tab      = split ( /\|/, $hash );
	my $checksum = $tab[0];

	return $checksum;
}
#######################################################
# get file_type from database
sub file_type ($) {
	my $name = shift (@_);    # file name

	# see sys/stat.h
	my $S_IFIFO  = 0010000;
	my $S_IFCHR  = 0020000;
	my $S_IFDIR  = 0040000;
	my $S_IFBLK  = 0060000;
	my $S_IFREG  = 0100000;
	my $S_IFLNK  = 0120000;
	my $S_IFSOCK = 0140000;

	# $Hashfile for changed file
	# Newval in compare mode for new files
	my $hash = $Hashfile{$name} || $Newval{$name};
	my @tab  = split ( /\|/, $hash );
	my $perm = $tab[3];

	# order is important !
	if ( ( $perm & $S_IFSOCK ) == $S_IFSOCK ) {
		return "socket";
	}
	elsif ( ( $perm & $S_IFLNK ) == $S_IFLNK ) {
		return "symbolic link";
	}
	elsif ( ( $perm & $S_IFREG ) == $S_IFREG ) {
		return "file";
	}
	elsif ( ( $perm & $S_IFBLK ) == $S_IFBLK ) {
		return "block device";
	}
	elsif ( ( $perm & $S_IFDIR ) == $S_IFDIR ) {
		return "directory";
	}
	elsif ( ( $perm & $S_IFCHR ) == $S_IFCHR ) {
		return "character device";
	}
	elsif ( ( $perm & $S_IFIFO ) == $S_IFIFO ) {
		return "fifo";
	}
	else {
		return "unknown type";
	}

}
#######################################################
# used to send "normal" messages
sub report($) {
	no strict 'refs';
	print $Report_url @_ if ( $Report_url ne 'null' );
}
#######################################################
# for info purpose
sub progress($) {
	print "progress @_\n";
}
#######################################################
# for debug purpose
sub debug($) {
	print "DEBUG: @_" if ($Verbose);
}
#######################################################
sub warning($) {
	my $text = shift (@_);
	warn "WARNING: $text";
}

#######################################################
# remove_trailing blanks
sub remove_trailing_spaces ($) {
	my $val = shift (@_);
	$$val =~ s/^\s+//;
	$$val =~ s/\s+$//;
}
#######################################################
# remove all spaces
sub remove_all_spaces ($) {
	my $val = shift (@_);
	$$val =~ s/\s+//g;
}
#######################################################
# test_double
# to detect bad configuration
sub test_double($) {
	my $elem = shift (@_);
	return exists $Toscan{$elem};
}
#######################################################
# used to exclude suffixes
sub test_exclude_suffix($) {
	my $fic = shift (@_);
	foreach my $sufx (@ExSufx) {
		if ( $fic =~ m/\.$sufx$/ ) {
			debug("find excluded suffix $sufx in $fic\n");
			return 1;
		}
	}
	return 0;
}
#######################################################
# expand_line
# expand config line : treat jokers, and split into <file attribute flagdir>
sub trait_line ($) {
	my $line = shift (@_);

	my $name;    # file name
	my $attribute = "";    # what to do on $name

	# for file with blank caracters
	if ( $line =~ m/^"(.*)"(.*)/ ) {
		$name      = $1;
		$attribute = $2;
	}
	else {
		( $name, $attribute ) = split ( " ", $line, 2 );
	}

	# for compatibilite with old syntaxe
	$attribute = $default_alias if ( !$attribute );

	# debug( "$name -> $attribute");

	my @tab;
	my $flagdir;

	# expand jokers if exists
	if ( $name =~ m/[\*\?]/ ) {
		@tab = glob($name);
		debug("expand $name in @tab\n");
	}
	elsif ( $name eq "/" ) {
		$flagdir = 1;
		push ( @tab, $name );
	}
	elsif ( $name =~ m?/$? ) {

		# remove trailing /
		$name =~ s?/$??;
		$flagdir = 1;
		push ( @tab, $name );
	}
	else {
		$flagdir = 0;
		push ( @tab, $name );
	}
	push ( @tab, $attribute, $flagdir );
	return @tab;
}
#######################################################
#resolv_alias
# expand and resolv action pattern, for alias definitions and files config
# be careful : not recursif
sub resolv_alias($$) {
	my $nline = shift (@_);    # line number in config file
	my $val   = shift (@_);    # value

	my $realval = "";

	remove_all_spaces( \$val );
	$val =~ m/^([^+-]*)/;      # first motif before any +/-
	my $first_alias = $1;
	debug("first alias : $first_alias\n");
	if ( !exists $alias{$first_alias} ) {

		# error : pattern not found
		warning(
"error in config file (line $nline : $val) : first alias \"$first_alias\" can not be resolved (ignored)\n"
		);
	}
	else {
		$realval = $alias{$first_alias};
	}

	# then other flags
	# first negative ones
	while ( $val =~ m/-([^+-]*)/g ) {
		my $remov_alias = $1;
		debug("soustractive alias $remov_alias\n");
		if ( !exists $alias{$remov_alias} ) {

			# remove bad aliases
			warning(
"error in config file (line $nline : $val) :  soustractive alias \"$remov_alias\" can not be resolved (ignored)\n"
			);
		}
		else {
			$realval =~ s/$remov_alias//;
		}
	}

	# then additives ones
	while ( $val =~ m/\+([^+-]*)/g ) {
		my $add_alias = $1;
		debug("additive alias $add_alias\n");
		if ( !exists $alias{$add_alias} ) {

			# remove bad aliases
			warning(
"error in config file (line $nline : $val) : additive alias \"$add_alias\" can not be resolved (ignored)\n"
			);
			next;
		}
		elsif ( $realval =~ m/$alias{$add_alias}/ ) {

			# do not allow duplicates
			warning(
"error in config file (line $nline : $val) : duplicate alias $add_alias in $realval\n"
			);
		}
		else {
			$realval .= $alias{$add_alias};
		}
	}

	if ( !$realval ) {
		$realval = $alias{$default_alias};
		warning(
			"set to default pattern \"$default_alias\" in config line line $.\n"
		);
	}
	return $realval;

}
#######################################################
# decode
# used to read directives : yes/1/true or no/0/false

sub decode ($$$) {
	my $val = lc( shift (@_) );    # nocase
	my $line   = shift (@_);       # line
	my $nuline = shift (@_);       # number of line

	if ( ( $val eq "yes" ) or ( $val eq "true" ) or ( $val eq "1" ) ) {
		return 1;
	}
	elsif ( ( $val eq "no" ) or ( $val eq "false" ) or ( $val eq "0" ) ) {
		return 0;
	}
	else {
		warning("bad value $val in config line line $nuline : $line\n");

		# return as no
		return 0;
	}
}
#######################################################
# get_names
# used by auto_control to know which program names to control
sub get_names() {
	my ( $name, $path, undef ) = fileparse($0);

	my @list;

	# because of links from afick.pl to afick
	my $afick = $path . "afick.pl";
	if ( -x $afick ) {
		push ( @list, $afick );
	}
	else {
		push ( @list, $0 );
	}

	my $aficktk = $path . "afick-tk.pl";

	# is afick-tk installed ?
	if ( -x $aficktk ) {
		push ( @list, $aficktk );
	}

	return @list;
}
#######################################################
# ignore_directive
# used by read_config for a common message
sub ignore_directive($$) {
	my $line    = shift (@_);
	my $numline = shift (@_);
	debug("ignore $line (line $numline) : overload by command line option\n");
}
#######################################################
# read_config :
# read the config file to build a list of directory to scan

sub read_config($) {
	my $config = shift (@_);    # config file name

	my @liste_toscan;

	debug("-------- begin of config file -------------\n");

	open( INIT, "<", $config )
	  or die "abort : can not open config file $config : $!\n";

	# for multi_line exclude_suffix : to detect first call
	my $first_exclude = 1;
	while ( my $line = <INIT> ) {
		chomp $line;

		$line =~ s/#.*//;    # skip comments
		remove_trailing_spaces( \$line );
		$line =~ s?\\?/?g;    # replace \ by / (for windows)

		next unless length($line);    # skip blank lines

		#debug( "config line = $line\n");

		if ( $line =~ m/\@\@/ ) {

			# skip macros lines
			###################
			debug("skipped macro in line $. : $line\n");
			next;
		}
		elsif ( $line =~ m/^=/ ) {

			# begin with a = : a directory to scan without recurse
			######################################################
			$line =~ s/^=\s*//;

			my @tab     = trait_line($line);
			my $flagdir = pop (@tab);
			my $masqb   = pop (@tab);
			my $masq    = resolv_alias( $., $masqb );

			foreach my $elem (@tab) {
				if ( test_double($elem) ) {
					warn
"WARNING: skipped config file $line (line $.) : $elem already in config file\n";
				}
				elsif ( -d $elem ) {
					push ( @liste_toscan, $elem );
					$Toscan{$elem} = $masq;
					if ($flagdir) {
						$Onlydir{$elem} = 1;
						debug("toscan without sub-dir : $elem masq $masq\n");
					}
					else {
						$Onlythis{$elem} = 1;
						debug("toscan only this inode : $elem masq $masq\n");
					}
				}
				elsif ( -e _ ) {
					warn
"WARNING: skipped config file $elem (line $.) : $elem not a directory\n";
				}
				else {
					warn
"WARNING: skipped config file $line (line $.) : $elem does not exists\n";
				}
			}    # foreach my $elem
		}
		elsif ( $line =~ m/:=/ ) {

			# a configuration line
			######################
			my ( $key, $val ) = split ( /:=/, $line, 2 );
			remove_trailing_spaces( \$val );

			if ( $key =~ m/^database/ ) {
				debug("found database : $val\n");
				$database = $val;
			}
			elsif ( $key =~ m/^verbose/ ) {
				if ( !defined $Verbose ) {
					debug("found verbose : $val\n");
					$Verbose = decode( $val, $line, $. );
				}
				else {
					ignore_directive( $line, $. );
				}
			}
			elsif ( $key =~ m/^warn_dead_symlinks/ ) {
				if ( !defined $Warn_dead_symlinks ) {
					debug("found warn_dead_symlinks : $val\n");
					$Warn_dead_symlinks = decode( $val, $line, $. );
				}
				else {
					ignore_directive( $line, $. );
				}
			}
			elsif ( $key =~ m/^report_full_newdel/ ) {
				if ( !defined $Report_full_newdel ) {
					debug("found report_full_newdel : $val\n");
					$Report_full_newdel = decode( $val, $line, $. );
				}
				else {
					ignore_directive( $line, $. );
				}
			}
			elsif ( $key =~ m/^warn_missing_file/ ) {
				if ( !defined $Warn_missing_file ) {
					debug("found warn_missing_file : $val\n");
					$Warn_missing_file = decode( $val, $line, $. );
				}
				else {
					ignore_directive( $line, $. );
				}

			}
			elsif ( $key =~ m/^report_url/ ) {
				debug("found report_url : $val\n");
				if ( lc($val) eq "stdout" ) {
					$Report_url = "stdout";
				}
				elsif ( lc($val) eq "stderr" ) {
					$Report_url = "stderr";
				}
				elsif ( lc($val) eq "null" ) {
					$Report_url = 'null';
				}
				else {
					warn
"WARNING: skipped config file $line (line $.) : bad value $val\n";
				}
			}
			elsif ( $key =~ m/^exclude_suffix/ ) {
				if ( $first_exclude && defined @ExSufx ) {
					ignore_directive( $line, $. );
				}
				else {
					$first_exclude = 0;
					debug("found exclude_suffix : $val\n");

					# enable multi-line syntaxe
					push ( @ExSufx, split ( ' ', $val ) );
				}
			}
			elsif ( $key =~ m/^timing/ ) {
				if ( !defined $Timing ) {
					debug("found timing : $val\n");
					$Timing = decode( $val, $line, $. );
				}
				else {
					ignore_directive( $line, $. );
				}
			}
			elsif ( $key =~ m/^running_files/ ) {
				if ( !defined $Running ) {
					debug("found running_files : $val\n");
					$Running = decode( $val, $line, $. );
				}
				else {
					ignore_directive( $line, $. );
				}
			}
			else {
				warn
"WARNING: skipped config file $line (line $.) : unknown directive $key\n";
			}
		}
		elsif ( $line =~ m/=/ ) {

			# is an alias definition
			########################
			remove_all_spaces( \$line );
			my ( $key, $val ) = split ( /=/, $line );
			my $realval = resolv_alias( $., $val );
			$alias{$key} = $realval;
			debug("alias $key -> $realval\n");
		}

		elsif ( $line =~ m/^!/ ) {

			# begin with ! : it is an exception
			###################################
			$line =~ s/^!\s*//;

			my @tab     = trait_line($line);
			my $flagdir = pop (@tab);          # and ignore
			my $masq    = pop (@tab);          # and ignore
			foreach my $elem (@tab) {
				if ( test_double($elem) ) {
					warn
"WARNING: skipped config file $line (line $.) : exception $elem already seen in config file\n";
				}
				else {
					debug("exception : $elem\n");
					$Toscan{$elem} = 0;
				}
			}
		}
		else {

			# directory or file to scan
			###########################
			my @tab     = trait_line($line);
			my $flagdir = pop (@tab);
			my $masqb   = pop (@tab);
			my $masq    = resolv_alias( $., $masqb );

			foreach my $elem (@tab) {
				if ( test_double($elem) ) {
					warn
"WARNING: skipped config file $line (line $.) : $elem already in config file\n";
				}
				elsif ( -e $elem ) {
					push ( @liste_toscan, $elem );
					debug("toscan : $elem masq $masq\n");
					$Toscan{$elem} = $masq;
				}
				elsif ($Warn_missing_file) {
					warn
"WARNING: skipped config file $line (line $.) : $elem does not exists\n";
				}
			}    # foreach my $elem
		}
	}    # foreach my $line
	close(INIT);

	# set default values if not founds
	$Verbose            = 0 if ( !defined $Verbose );
	$Warn_dead_symlinks = 0 if ( !defined $Warn_dead_symlinks );
	$Report_full_newdel = 0 if ( !defined $Report_full_newdel );
	$Warn_missing_file  = 0 if ( !defined $Warn_missing_file );
	$Timing             = 0 if ( !defined $Timing );
	$Running            = 0 if ( !defined $Running );

# add some files for internal check : this program, the graphical interface, the config file
#my $masq = resolv_alias( $., "all" );
	my $masq = $alias{"all"};

	my @list = get_names();

	foreach my $elem (@list) {
		$Toscan{$elem} = $masq;
		push ( @liste_toscan, $elem );
	}

	$Toscan{$config} = $masq;
	push ( @liste_toscan, $config );

	# a summary of used config
	print_config();

	debug("-------- end of config file -------------\n");

	return @liste_toscan;
}
#######################################################
# display
# convert fields in a human way
sub display($@) {
	my $elem = shift (@_);
	my @info = @_;

	if ( $elem == 1 ) {

		# device is in hexa, ex 309 == major 3, minor 9 == /dev/hda9
		return sprintf "%lx", $info[$elem];
	}
	elsif ( $elem == 3 ) {

		# perm is in octal
		return sprintf "%lo", $info[$elem];
	}
	elsif ( $elem >= 9 ) {

		# dates
		if ( $info[$elem] ) {
			return gmtime( $info[$elem] );
		}
		else {
			return 0;
		}
	}
	else {
		return $info[$elem];
	}
}

#######################################################
# display_all
# convert in a human way all the record
sub display_all($) {
	my $record = shift (@_);
	my @fileinfo = split ( /\|/, $record );

	my $output;

	my $max = scalar @fileinfo;
	for ( my $i = 0 ; $i < $max ; $i++ ) {
		$output .= display( $i, @fileinfo ) . "|";
	}
	return $output;
}
#######################################################
# md5sum :
# build a MD5 checksum on the given file
sub md5sum($) {
	my $name = shift (@_);    # file name

	# !!! the Digest module is not in standard
	# so we can not write Digest->new("MD5");
	my $ctx = Digest::MD5->new();

	if ( open( FILE, "<", "$name" ) ) {
		binmode(FILE);
		$ctx->addfile(*FILE);
		close(FILE);
		my $sum = $ctx->b64digest;
		return $sum;
	}
	else {
		warning("(md5sum) can not open $name for checksum: $!\n");
		return 0;
	}
}
#######################################################
# file_info :
# build a complete info on the given file
sub file_info($$) {
	my $name = shift (@_);
	my $masq = shift (@_);

	# for afick-tk progress text
	progress("$name") if ($Progress);

	debug("$name|");

	my @fileinfo = lstat($name);

	my $sum;
	if ( -f $name ) {
		if ( ( $masq =~ m/h/ ) or ( $masq =~ m/5/ ) ) {
			$sum = md5sum($name);

			# prepare for other checksum
			#		} elsif ($masq =~ m/1/) {
			#			$sum = sha1sum($name);
		}
		else {
			$sum = 0;
		}
	}
	else {
		$sum = 0;
	}

	# Each file gets a pipe-delimited entry, with format as follows:
	#
	# field 0 = checksum hash
	# field 1 = device number of filesystem
	# field 2 = inode number
	# field 3 = file mode (type/permissions)
	# field 4 = number of hard links to file
	# field 5 = uid of the file
	# field 6 = gid of the file
	# field 7 = total size of file, in bytes
	# field 8 = actual number of blocks allocated
	# field 9 = last access time in seconds since the epoch
	# field 10 = last modify time (since epoch)
	# field 11 =  inode change time
	#
	# See the perl "stat" command or man stat() for more information.

	# default value is 0
	my (
		$device, $inode, $perm,  $link,  $uid, $gid,
		$size,   $bloc,  $atime, $mtime, $ctime
	  )
	  = ( 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 );
	$device = $fileinfo[0]  if ( $masq =~ m/d/ );
	$inode  = $fileinfo[1]  if ( $masq =~ m/i/ );
	$perm   = $fileinfo[2]  if ( $masq =~ m/p/ );
	$link   = $fileinfo[3]  if ( $masq =~ m/n/ );
	$uid    = $fileinfo[4]  if ( $masq =~ m/u/ );
	$gid    = $fileinfo[5]  if ( $masq =~ m/g/ );
	$size   = $fileinfo[7]  if ( $masq =~ m/s/ );
	$atime  = $fileinfo[8]  if ( $masq =~ m/a/ );
	$mtime  = $fileinfo[9]  if ( $masq =~ m/m/ );
	$ctime  = $fileinfo[10] if ( $masq =~ m/c/ );
	$bloc   = $fileinfo[12] if ( $masq =~ m/b/ );

	my $entry =
"$sum|$device|$inode|$perm|$link|$uid|$gid|$size|$bloc|$atime|$mtime|$ctime";

	debug( display_all($entry) . "\n" );

	if ($Running) {
		warning("$name as been modified during the program run\n")
		  if ( ( $masq =~ m/m/ ) and ( -M _ < 0 ) );
		warning("inode of $name as been modified during the program run\n")
		  if ( ( $masq =~ m/c/ ) and ( -C _ < 0 ) );
	}

	return $entry;
}
#######################################################
# compare_entry :
# compare 2 file infos and print the difference
sub compare_entry($@@) {
	my $name   = shift (@_);
	my $entry1 = shift (@_);    # old value
	my $entry2 = shift (@_);    # new value

	my @fileinfo1 = split ( /\|/, $entry1 );
	my @fileinfo2 = split ( /\|/, $entry2 );

	my @field = (
		'MD5       ',           # 0
		'device    ',           # 1
		'inode     ',           # 2
		'file mode ',           # 3
		'links     ',           # 4
		'uid       ',           # 5
		'gid       ',           # 6
		'filesize  ',           # 7
		'blocs     ',           # 8
		'atime     ',           # 9
		'mtime     ',           # 10
		'ctime     '            # 11
	);

	# Each file gets a pipe-delimited entry, with format as follows:
	my $i;
	my $max = scalar @fileinfo1;
	for ( $i = 0 ; $i < $max ; $i++ ) {
		if ( $fileinfo1[$i] ne $fileinfo2[$i] ) {
			report( "\t$field[$i]\t : "
				  . display( $i, @fileinfo1 ) . "\t"
				  . display( $i, @fileinfo2 )
				  . "\n" );
		}                       # if fileinfo ne
	}    # for
}

#######################################################
# is_exception :
# return true if the file name belong exception list
sub is_exception($) {
	my $elem = shift (@_);
	return ( exists( $Toscan{$elem} ) && ( $Toscan{$elem} eq "0" ) );
}
#######################################################
# parcours :
# the general sub to scan a directory (recursive)
# should be called by config list entries
sub parcours($$$$) {
	my $rep     = shift (@_);    # file or directory to scan
	my $traite  = shift (@_);    # sub to call
	my $masq    = shift (@_);    # herited masq
	my $newflag = shift (@_);    # herited new flag

	#debug ("parcours : $rep $masq\n");

	my @liste;

	# first build a list of file
	if ( -d $rep ) {
		return if ( $rep =~ /\/\.\.?$/ );

		if ( exists $Toscan{$rep} ) {
			$masq = $Toscan{$rep};

			debug("change masq to $masq for $rep\n");
		}
		$traite->( $rep, $masq, $newflag );

		# equal : do not scan into the directory
		return if ( exists $Onlythis{$rep} );

		if ( !-l $rep ) {

			if ( !opendir( DIR, $rep ) ) {
				warning("(parcours) can not open directory $rep : $!\n");
				return;
			}
			my $elem;
			while ( defined( $elem = readdir(DIR) ) ) {
				next if $elem =~ /^\.\.?$/;

				#debug( "readdir : $elem");

				# equal : do not scan sub-dir
				next
				  if (  ( exists $Onlydir{$rep} )
					and ( -d $rep . "/" . $elem ) );

				# special case : / to avoid //
				if ( $rep eq "/" ) {
					push ( @liste, '/' . $elem );
				}
				else {
					push ( @liste, $rep . '/' . $elem );
				}
			}
			closedir(DIR);
		}
		else {

			# on ne suit pas les liens sur les repertoires
			debug("$rep skipped (link on directory)\n");
		}
	}
	else {
		push ( @liste, $rep );
	}

	parcours2( $rep, $traite, $masq, $Newfiles{$rep}, @liste );
}
#######################################################

sub parcours2($$@) {
	my $rep     = shift (@_);    # file or directory to scan
	my $traite  = shift (@_);    # sub to call
	my $masq    = shift (@_);    # herited masq
	my $newflag = shift (@_);    # herited new flag
	my @liste   = @_;

	foreach my $fic (@liste) {
		next if $fic =~ /^\.\.?$/;
		next if ( is_exception($fic) );
		next if ( test_exclude_suffix($fic) );

		# heritage, set to all by default (ex : -l action)
		my $lmasq = $masq || $alias{"all"};

		# surcharge
		if ( exists $Toscan{$fic} ) {
			$lmasq = $Toscan{$fic};

			#debug ("change masq to $lmasq for $fic");
		}

		debug("parcours : $fic\n");

		if ( -d $fic ) {
			parcours( $fic, $traite, $lmasq, $newflag );
		}    # directory
		elsif ( -l $fic ) {
			my $whatlink = readlink($fic);

			# is the link an absolute path ?
			if ( !( $whatlink =~ m/^\// ) ) {
				$whatlink = $rep . "/" . $whatlink;
			}
			if ( !-e $whatlink ) {

				$Dangling{$fic} = $whatlink;
				next;
			}
			else {
				$traite->( $fic, $lmasq, $Newfiles{$rep} );
			}
		}    # link
		else {
			$traite->( $fic, $lmasq, $Newfiles{$rep} );
		}    # normal file
	}    # foreach
}
#######################################################
# wanted_create
# process a file in create mode
sub wanted_create($$) {
	my $name = shift (@_);
	my $masq = shift (@_);

	$Finded{$name} = 1;

	my $entry = file_info( $name, $masq );

	$Hashfile{$name} = $entry;
}
#######################################################
# end_report
sub end_report() {

	# Generate a MD5 for the dbm file.
	my $sum = md5sum( $database . $database_ext );

	report( "-" x 65 );
	report("\nMD5 hash of $database => $sum\n\n");
	write_control($sum);
}
#######################################################
# create (base liste)
# create a new database (empty the existing one if exists)
sub create($@) {
	my $database = shift (@_);
	my @listerep = @_;
	tie( %Hashfile, $Dbm, $database, O_WRONLY | O_CREAT, 0666 )
	  or die "abort (create) : can not open $Dbm $database : $!";

	# delete all previous keys
	%Hashfile = ();

	foreach my $elem (@listerep) {

		# to skip already scanned files
		next if ( exists $Finded{$elem} );

		parcours( $elem, \&wanted_create, "", "" );
	}
	my $nbscan = scalar( keys %Finded );
	untie %Hashfile;
	print_dangling();

	report("\nHash database created successfully. $nbscan files entered.\n");

	end_report();
}
#######################################################
# wanted_update
# process a file in update mode
sub wanted_update($$) {
	my $name    = shift (@_);
	my $masq    = shift (@_);
	my $newflag = shift (@_);

	$Finded{$name} = 1;

	my $entry = file_info( $name, $masq );

	if ( $Hashfile{$name} ) {

		# test for changes
		my $old_entry = $Hashfile{$name};
		if ( $old_entry ne $entry ) {

			# keep values to compare
			$Oldval{$name} = $old_entry;
			$Newval{$name} = $entry;
			$Nbmod++;
			$Hashfile{$name} = $entry if ($Update);
		}
	}
	else {

		# new entry
		if ( defined $newflag ) {
			$Newfiles{$name} = $newflag + 1;
		}
		else {
			$Newfiles{$name} = 1;
		}

		$Newval{$name} = $entry;

		#debug ("$name heritage : $newflag value $Newfiles{$name}\n");
		$Hashfile{$name} = $entry if ($Update);
	}
}    # end of wanted
#######################################################
# update :
# compare or update a database with the system
sub update($@) {
	my $database = shift (@_);
	my @listerep = @_;
	tie( %Hashfile, $Dbm, $database, O_RDWR, 0666 )
	  or die "abort (update) : can not open $Dbm $database : $!";

	calc_control();

	# for afick-tk progress bar
	# guess the number of file does not change too much
	my $total = scalar( keys %Hashfile );
	progress("total $total") if ($Progress);

	# scan file list
	foreach my $elem (@listerep) {

		# skip already scanned
		next if ( exists $Finded{$elem} );

		parcours( $elem, \&wanted_update, "", "" );
	}

	# analysis
	my $nbchange = $Nbmod;
	my $nbdelete = 0;
	my $nbnew    = 0;
	if ( !$File ) {

		# look for new files
		my $nb = 0;
		foreach my $key ( sort keys %Newfiles ) {

			# report only if asked all ($Report_full_newdel) or first level
			if ($Report_full_newdel) {
				report( "new " . file_type($key) . " : $key\n" );
			}
			elsif ( $Newfiles{$key} == 1 ) {
				if ($nb) {
					report("  number of new files : $nb\n");
				}
				report( "new " . file_type($key) . " : $key\n" );
				$nb = 0;
			}
			else {
				$nb++;
			}

			$Nbmod++;
			$nbnew++;
		}
		if ($nb) {
			report("  number of new files : $nb\n");
		}

		# look for deleted files
		my %deleted;
		$nb = 0;
		foreach my $key ( sort keys %Hashfile ) {
			if ( !exists( $Finded{$key} ) ) {
				if ($Report_full_newdel) {
					report( "deleted " . file_type($key) . " : $key\n" );
					$nb = 0;
				}
				else {

				  # get directory for a file or parent directory for a directory
					my $path = dirname($key);

					#print "--- deleted $key $path\n";
					if ( exists $deleted{$path} ) {
						$deleted{$key} = $deleted{$path} + 1;
						$nb++;
					}
					else {
						$deleted{$key} = 1;
						if ($nb) {
							report("  number of deleted files : $nb\n");
						}
						$nb = 0;
						report( "deleted " . file_type($key) . " : $key\n" );
					}
				}
				delete $Hashfile{$key} if ($Update);
				$Nbmod++;
				$nbdelete++;
			}
		}
		if ($nb) {
			report("  number of deleted files : $nb\n");
		}
	}
	print_changed(0);
	print_dangling();

	report("\ndetailed changes\n") if $nbchange;
	print_changed(1);

	untie %Hashfile;

	if ($Update) {
		report("\nHash database updated successfully : ");
	}
	else {
		report("\nHash database : ");
	}
	my $nbscan   = scalar( keys %Finded );
	my $dangling = scalar( keys %Dangling );
	report(
"$nbscan files scanned, $Nbmod changed (new : $nbnew; delete : $nbdelete; changed : $nbchange; dangling : $dangling)\n"
	);

	end_report();

	# return a status
	return oct(
		join ( '',
			'0b',
			map { $_ ? 1 : 0; } ( $nbnew, $nbdelete, $nbchange, $dangling ) )
	);
}

#######################################################
# print_dangling
# print all danglink symbolic links
sub print_dangling() {
	return unless ($Warn_dead_symlinks);
	foreach my $key ( keys %Dangling ) {
		report("Dangling link : [$key] => [$Dangling{$key}]\n");
	}
}

#######################################################
# print_dbm
# print the database content
sub print_dbm($) {
	my $database = shift (@_);
	tie( %Hashfile, $Dbm, $database, O_RDONLY, 0666 )
	  or die "abort (print_dbm) : Couldn't tie $Dbm file $database : $!";

	calc_control();

	foreach my $key ( sort keys %Hashfile ) {
		report( file_type($key) . " $key|"
			  . display_all( $Hashfile{$key} )
			  . "\n" );
	}
	report( "number of file : " . scalar( keys %Hashfile ) . "\n" );
	untie %Hashfile;
}
#######################################################
# print a summary or detailed info on changes
sub print_changed($) {
	my $detailed = shift (@_);
	foreach my $key ( sort keys %Oldval ) {
		report( "changed " . file_type($key) . " : $key\n" );
		compare_entry( $key, $Oldval{$key}, $Newval{$key} ) if ($detailed);
	}
}
#######################################################
# print internal flag state
sub print_config() {
	debug("report_url=$Report_url\n");
	debug("Verbose=$Verbose\n");
	debug("Warn_dead_symlinks=$Warn_dead_symlinks\n");
	debug("Report_full_newdel=$Report_full_newdel\n");
	debug("Warn_missing_file=$Warn_missing_file\n");
	debug("exclude_suffix=@ExSufx\n");

	#debug("Output_format=$Output_format\n");
}
#######################################################
# usage
# print some help
sub usage($) {
	my $version = shift (@_);
	print <<EOHELP;

Usage: $0 [mandatory action] [other options]

Mandatory action (one and only one must be used) : 
 -i|--init                    initialize the hash.dbm database
 -k|--compare                 compare the hash.dbm database
 -l|--list fic1 .. fic2       check the files given in arg 
 -u|--update                  compare and update the hash.dbm database
 -p|--print                   print content of dbm database

Other options
 -c|--config_file file        configname of config file to use
 -f|--full_newdel             report full information for new or deleted directories
                               default: no
			       reverse : --nofull_newdel 
 -m|--missing_files           warn about files declared in config files 
                               which do not exists, 
			       default: no; 
			       reverse : --nomissing_files
 -r|--running_files           warn about "running" files : modified since program begin
                               default: no 
                               reverse: --norunning_files
 -s|--dead_symlinks           warn about dead symlinks 
                               default: no 
                               reverse: --nodead_symlinks
 -t|--timing		      Print timing statistics
				default: no
				reverse : --notiming
 -v|--verbose                 toggle verbose mode; 
                               default: no 
 			       reverse : --noverbose
 -P|--progress		      display the name of scanned files, to be used only by afick-tk
 -h|--help                    show this help page
 -V|--version                 show afick version
 -x|--exclude_suffix ext1 ext2        list of suffixes to ignore

Disclaimer:
This script is intended to provide a means for
detecting changes made to files, via a regular
comparison of MD5 hashes to an established baseline. 

Copyright (c) 2002 Eric Gerbier <gerbier\@users.sourceforge.net>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
EOHELP

}

#######################################################
sub version($) {
	my $version = shift (@_);
	print "\n";
	print "afick : another file integrity checker\nversion $version\n";
}
#######################################################
# read database previous checksum and compare with new one
sub read_control($) {
	my $newchecksum  = shift (@_);
	my $control_file = $database . $control_ext;

	my $oldchecksum;
	if ( open( CONTROL, "<", $control_file ) ) {
		$oldchecksum = <CONTROL>;
		chomp($oldchecksum);
		close(CONTROL);
		if ( $oldchecksum ne $newchecksum ) {
			warn
"WARNING: internal change in afick database $database (see below)\n";
		}
	}
	else {
		warning(
			"(read_control) can not read control file $control_file : $!\n");
	}
}
#######################################################
# write database checksum for next run
sub write_control ($) {
	my $checksum     = shift (@_);                 # the string to write
	my $control_file = $database . $control_ext;

	# to avoid Tainted warnings
	$control_file = $1 if ( $control_file =~ /(.+)/ );

	if ( open( CONTROL, ">", $control_file ) ) {
		print CONTROL "$checksum";
		close(CONTROL);
	}
	else {
		warning(
			"(write_control) can not write in control file $control_file : $!\n"
		);
	}
}
#######################################################
# compare old and new checksum for afick main components
sub control ($) {
	my $name = shift (@_);    # file name to control

	my $newchecksum = md5sum($name);
	my $oldchecksum = get_md5($name);

	if ( !defined $oldchecksum ) {

		# first run
		return;
	}
	elsif ( $oldchecksum ne $newchecksum ) {
		warning("(control) afick internal change : $name (see below)\n");
	}
}
#######################################################
# check if afick change since last run
sub calc_control() {

	my @list = get_names();

	foreach my $elem (@list) {
		control($elem);    # afick.pl
	}

	control($configfile);    # afick-tk.pl

	my $database_check = md5sum( $database . $database_ext );
	read_control($database_check);

}
########################## main #############################

if ( $^O =~ m/linux/ ) {
	$default_config_file = "/etc/afick.conf";
}
elsif ( $^O =~ m/^MSWin/ ) {
	$default_config_file = "xp.conf";
}
else {
	$default_config_file = "afick.conf";
}

$| = 1;

Getopt::Long::Configure('no_ignore_case');
unless (
	GetOptions(
		"config_file|c=s"    => \$configfile,
		"full_newdel|f!"     => \$Report_full_newdel,
		"help|h"             => \$help,
		"init|i"             => \$init,
		"compare|k"          => \$compare,
		"list|l=s"           => \$list,
		"missing_files|m!"   => \$Warn_missing_file,
		"print|p"            => \$print,
		"progress|P"         => \$Progress,
		"running_files|r!"   => \$Running,
		"dead_symlinks|s!"   => \$Warn_dead_symlinks,
		"verbose|v!"         => \$Verbose,
		"version|V"          => \$Version,
		"exclude_suffix|x=s" => \$Sufx_list,
		"timing|t!"          => \$Timing,
		"update|u"           => \$Update
	)
  )
{
	usage($version);
	die "abort : incorrect option\n";
}

# -h : help
if ($help) {
	usage($version);
	exit;
}

# -V : version
if ($Version) {
	version($version);
	exit;
}

# compute exclude suffix list from arg line
if ($Sufx_list) {
	@ExSufx = split ( ? ?, $Sufx_list );
}

if ($configfile) {
	debug("config file : $configfile\n");
}
elsif ( -e $default_config_file ) {
	$configfile = $default_config_file;
}
else {
	usage($version);
	die
"abort : missing configfile name (-c flag) and default config file $default_config_file\n";
}
@listerep = read_config($configfile);

# no we should have a database name
if ( !$database ) {
	usage($version);
	die
"abort : missing database name in flag list (-b) and in config file $configfile\n";
}

my $return_value = 0;

# actions
if ($list) {

	# -l : liste
	#@listerep = trait_config( split ( ? ?, $liste ) );
	@listerep     = split ( ? ?, $list );
	$File         = 1;
	$return_value = update( $database, @listerep );
}
elsif ($print) {

	# -p : print
	print_dbm($database);
}
elsif ($init) {

	# -i : init
	create( $database, @listerep );
}
elsif ($Update) {

	# -u : update
	$return_value = update( $database, @listerep );
}
elsif ($compare) {

	# -k : check
	$return_value = update( $database, @listerep );
}
else {
	usage($version);
	die "abort : no action to do (-i, -u, -k, -l, -p) \n";
}

# timing info
if ($Timing) {
	my ( $user, $system, $cuser, $csystem ) = times();
	print "user time : $user; system time : $system\n";
}

exit $return_value;

__END__

=head1 NAME

afick : Another File Integrity Checker

=head1 SYNOPSIS

=head1 OPTIONS

=head1 DESCRIPTION

=cut
